import os
import pandas as pd
import pingouin as pg
import seaborn as sns
import matplotlib.pyplot as plt

from config import VALUE_COLUMN, HISTOGRAM_BINS, OUTPUT_PATH
from study1.base_data_processing import get_questionnaire_data
from study1.data_processing import get_errors, get_times
from study1.s1_config import PRELIMINARY_CHECK_PATH, NORMALITY_IMAGES_PATH, ANOVA_PATH, STUDY_KEY, QUESTIONNAIRE_PATH


def check_preliminaries(df: pd.DataFrame, df_type: str,
                        calc_norm: bool = True, calc_homo: bool = True, create_norm_img: bool = True):
    os.makedirs(PRELIMINARY_CHECK_PATH, exist_ok=True)
    os.makedirs(NORMALITY_IMAGES_PATH, exist_ok=True)

    if df_type is "E":
        value_types = ["NUMERIC", "BOOL"]
    else:
        value_types = ["TIME"]

    df_grouped = df.groupby(by=["questionType", "fileNr", "background", "valueType"]).mean().reset_index()

    for value_type in value_types:
        df_grouped_vt = df_grouped[df_grouped.valueType == value_type]
        df_vt = df[df.valueType == value_type]

        # Normality check
        if calc_norm:
            n = pg.normality(dv=VALUE_COLUMN, group="background", data=df_grouped_vt)
            n.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Norm) - background.csv")
            n = pg.normality(dv=VALUE_COLUMN, group="questionType", data=df_grouped_vt)
            n.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Norm) - questionType.csv")
            n = pg.normality(dv=VALUE_COLUMN, group="complexity", data=df_vt)
            n.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Norm) - complexity.csv")
        # Homoscedasticity check
        if calc_homo:
            h = pg.homoscedasticity(dv=VALUE_COLUMN, group="background", method="levene", data=df_grouped_vt)
            h.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Homo) - background.csv")
            h = pg.homoscedasticity(dv=VALUE_COLUMN, group="questionType", method="levene", data=df_grouped_vt)
            h.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Homo) - questionType.csv")
            h = pg.homoscedasticity(dv=VALUE_COLUMN, group="complexity", method="levene", data=df_vt)
            h.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Homo) - complexity.csv")

        # Normality check with images
        if create_norm_img:
            sns.distplot(df_vt[VALUE_COLUMN], bins=HISTOGRAM_BINS * 5, kde=True, rug=True)
            plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - Complete")
            plt.close()
            for background in df_vt.background.unique():
                sns.distplot(df_vt[df_vt.background == background][VALUE_COLUMN], bins=HISTOGRAM_BINS, kde=True, rug=True)
                plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - BG = {background}")
                plt.close()
            for question_type in df_vt.questionType.unique():
                sns.distplot(df_vt[df_vt.questionType == question_type][VALUE_COLUMN], bins=HISTOGRAM_BINS, kde=True, rug=True)
                plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - QT = {question_type}")
                plt.close()
            for complexity in df_vt.complexity.unique():
                sns.distplot(df_vt[df_vt.complexity == complexity][VALUE_COLUMN], bins=HISTOGRAM_BINS, kde=True, rug=True)
                plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - CX = {complexity}")
                plt.close()


def statistical_tests(error_df: pd.DataFrame, times_df: pd.DataFrame):
    os.makedirs(ANOVA_PATH, exist_ok=True)

    # --- H1 ------------------------------------------------
    e_df_grouped1 = error_df.groupby(by=["questionType", "fileNr", "background", "valueType"]).mean().reset_index()
    t_df_grouped1 = times_df.groupby(by=["questionType", "fileNr", "background", "valueType"]).mean().reset_index()

    pg.anova(dv=VALUE_COLUMN, between="background", data=e_df_grouped1[e_df_grouped1.valueType == "NUMERIC"]).to_csv(
               path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [NUMERIC] Anova - background.csv")
    pg.anova(dv=VALUE_COLUMN, between="background", data=e_df_grouped1[e_df_grouped1.valueType == "BOOL"]).to_csv(
               path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [BOOL] Anova - background.csv")
    pg.anova(dv=VALUE_COLUMN, between="background", data=t_df_grouped1[t_df_grouped1.valueType == "TIME"]).to_csv(
               path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [TIME] Anova - background.csv")

    # --- H2 ------------------------------------------------
    pg.rm_anova(dv=VALUE_COLUMN, within="complexity", subject="fileNr", correction=True, data=error_df[error_df.valueType == "NUMERIC"], detailed=True).to_csv(
               path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [NUMERIC] Rm_Ano - complexity.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within="complexity", subject="fileNr", correction=True, data=error_df[error_df.valueType == "BOOL"], detailed=True).to_csv(
               path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [BOOL] Rm_Ano - complexity.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within="complexity", subject="fileNr", correction=True, data=times_df[times_df.valueType == "TIME"], detailed=True).to_csv(
               path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [TIME] Rm_Ano - complexity.csv")

    pg.pairwise_ttests(dv=VALUE_COLUMN, within="complexity", subject="fileNr", padjust="fdr_bh", tail="one-sided", data=error_df[error_df.valueType == "NUMERIC"], return_desc=True).to_csv(
                       path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [NUMERIC] TTest - complexity.csv")
    pg.pairwise_ttests(dv=VALUE_COLUMN, within="complexity", subject="fileNr", padjust="fdr_bh", tail="one-sided", data=error_df[error_df.valueType == "BOOL"], return_desc=True).to_csv(
                       path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [BOOL] TTest - complexity.csv")

    # --- H3 ------------------------------------------------
    e_df_grouped2 = error_df.groupby(by=["questionType", "fileNr", "background", "complexity", "valueType"]).mean().reset_index()
    t_df_grouped2 = times_df.groupby(by=["questionType", "fileNr", "background", "complexity", "valueType"]).mean().reset_index()

    pg.mixed_anova(dv=VALUE_COLUMN, between="background", within="complexity", subject="fileNr", correction=True, data=e_df_grouped2[e_df_grouped2.valueType == "NUMERIC"]).to_csv(
                   path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [NUMERIC] Mix_An - background+complexity.csv")
    pg.mixed_anova(dv=VALUE_COLUMN, between="background", within="complexity", subject="fileNr", correction=True, data=e_df_grouped2[e_df_grouped2.valueType == "BOOL"]).to_csv(
                   path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [BOOL] Mix_An - background+complexity.csv")
    pg.mixed_anova(dv=VALUE_COLUMN, between="background", within="complexity", subject="fileNr", correction=True, data=t_df_grouped2[t_df_grouped2.valueType == "TIME"]).to_csv(
                   path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [TIME] Mix_An - background+complexity.csv")


# ------- H4 (+ Further Findings) ---------------------------
def questionnaire_tests(questionnaire_df: pd.DataFrame):
    os.makedirs(QUESTIONNAIRE_PATH, exist_ok=True)

    health_keys = ["fatigue", "concentration", "motivation", "headache", "dryEyes"]
    health_before_df = questionnaire_df[["fatigueBefore", "concentrationBefore", "motivationBefore", "headacheBefore", "dryEyesBefore"]]
    health_after_df = questionnaire_df[["fatigueAfter", "concentrationAfter", "motivationAfter", "headacheAfter", "dryEyesAfter"]]

    health_before_df.columns = health_keys
    health_after_df.columns = health_keys
    health_delta_df = health_after_df - health_before_df
    health_delta_df["fileNr"] = questionnaire_df.fileNr
    health_delta_df["background"] = questionnaire_df.background

    for key in health_keys:
        pg.kruskal(dv=key, between="background", data=health_delta_df).to_csv(
                   path_or_buf=f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} Krusk - h-{key} delta.csv")

    tlx_keys = [f"tlx{i}" for i in range(1, 7)]
    perception_keys = [f"{k}{i}" for i in range(1, 4) for k in ["generalQuestion", "generalPerception", "backgPerception"]]
    other_keys = ["height", "age"] + [f"preQuestion{i}" for i in range(1, 5)]

    for key in tlx_keys + perception_keys + other_keys:
        pg.kruskal(dv=key, between="background", data=questionnaire_df).to_csv(
                   path_or_buf=f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} Krusk - {key}.csv")

    sub_df = questionnaire_df.drop(labels=["sex", "fileNr"], axis=1)
    sub_df.mean().to_csv(f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} Mean.csv", header=True)
    sub_df.std().to_csv(f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} STD.csv", header=True)


if __name__ == '__main__':
    os.makedirs(OUTPUT_PATH, exist_ok=True)
    os.makedirs(PRELIMINARY_CHECK_PATH, exist_ok=True)

    error_df = get_errors(
        filter_methods=[
            lambda df: df[~df.questionType.isin(["Determine Range", "Extreme Values", "Delta", "Min", "Max"])],
        ]
    )
    times_df = get_times(
        filter_methods=[
            lambda df: df[(df.note.isin(["SolvingDuration"])) & (df.questionNr != 0)]
        ]
    )
    questionnaire_df = get_questionnaire_data()

    check_preliminaries(error_df, "E")
    check_preliminaries(times_df, "T")
    statistical_tests(error_df, times_df)
    questionnaire_tests(questionnaire_df)


