import os

import numpy as np
import pandas as pd
import pingouin as pg
import seaborn as sns
import matplotlib.pyplot as plt

from config import VALUE_COLUMN, HISTOGRAM_BINS, OUTPUT_PATH
from study2.base_data_processing import get_questionnaire_data
from study2.data_processing import calc_values, calc_time_values
from study2.s2_config import PRELIMINARY_CHECK_PATH, NORMALITY_IMAGES_PATH, STUDY_KEY, ANOVA_PATH, QUESTIONNAIRE_PATH, \
    QUESTION_TYPES_CALCULATED, QUESTION_TYPES


def check_preliminaries(df: pd.DataFrame, df_type: str,
                        calc_norm: bool = True, calc_homo: bool = True, create_norm_img: bool = True):
    os.makedirs(PRELIMINARY_CHECK_PATH, exist_ok=True)
    os.makedirs(NORMALITY_IMAGES_PATH, exist_ok=True)

    if df_type is "E":
        value_types = ["NUMERIC", "BOOL"]
    else:
        value_types = ["TIME"]

    for value_type in value_types:
        df_vt = df[df.valueType == value_type]

        # Normality check
        if calc_norm:
            n = pg.normality(dv=VALUE_COLUMN, group="background", data=df_vt)
            n.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Norm) - background.csv")
            n = pg.normality(dv=VALUE_COLUMN, group="questionType", data=df_vt)
            n.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Norm) - questionType.csv")
            n = pg.normality(dv=VALUE_COLUMN, group="focus", data=df_vt)
            n.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Norm) - focus.csv")
        # Homoscedasticity check
        if calc_homo:
            h = pg.homoscedasticity(dv=VALUE_COLUMN, group="background", method="levene", data=df_vt)
            h.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Homo) - background.csv")
            h = pg.homoscedasticity(dv=VALUE_COLUMN, group="questionType", method="levene", data=df_vt)
            h.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Homo) - questionType.csv")
            h = pg.homoscedasticity(dv=VALUE_COLUMN, group="focus", method="levene", data=df_vt)
            h.to_csv(f"{PRELIMINARY_CHECK_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (Homo) - focus.csv")

        # Normality check with images
        if create_norm_img:
            sns.distplot(df_vt[VALUE_COLUMN], bins=HISTOGRAM_BINS * 5, kde=True, rug=True)
            plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - Complete")
            plt.close()
            for background in df_vt.background.unique():
                sns.distplot(df_vt[df_vt.background == background][VALUE_COLUMN], bins=HISTOGRAM_BINS, kde=True, rug=True)
                plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - BG = {background}")
                plt.close()
            for question_type in df_vt.questionType.unique():
                sns.distplot(df_vt[df_vt.questionType == question_type][VALUE_COLUMN], bins=HISTOGRAM_BINS, kde=True, rug=True)
                plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - QT = {question_type}")
                plt.close()
            for focus in df_vt.focus.unique():
                sns.distplot(df_vt[df_vt.focus == focus][VALUE_COLUMN], bins=HISTOGRAM_BINS, kde=True, rug=True)
                plt.savefig(f"{NORMALITY_IMAGES_PATH}/{STUDY_KEY} [{df_type}] [{value_type}] (NImg) - FO = {focus}")
                plt.close()


def statistical_tests(error_df: pd.DataFrame, times_df: pd.DataFrame):
    os.makedirs(ANOVA_PATH, exist_ok=True)

    # --- H5 ------------------------------------------------
    pg.rm_anova(dv=VALUE_COLUMN, within=["background"], subject="fileNr", correction=True, data=error_df[error_df.valueType == "NUMERIC"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [NUMERIC] RM_Ano - background.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within=["background"], subject="fileNr", correction=True, data=error_df[error_df.valueType == "BOOL"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [BOOL] RM_Ano - background.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within=["background"], subject="fileNr", correction=True, data=times_df[times_df.valueType == "TIME"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [TIME] RM_Ano - background.csv")

    # --- H6 ------------------------------------------------
    pg.rm_anova(dv=VALUE_COLUMN, within=["focus"], subject="fileNr", correction=True, data=error_df[error_df.valueType == "NUMERIC"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [NUMERIC] RM_Ano - focus.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within=["focus"], subject="fileNr", correction=True, data=error_df[error_df.valueType == "BOOL"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [BOOL] RM_Ano - focus.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within=["focus"], subject="fileNr", correction=True, data=times_df[times_df.valueType == "TIME"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [TIME] RM_Ano - focus.csv")

    # --- H8 ------------------------------------------------
    pg.rm_anova(dv=VALUE_COLUMN, within=["background", "focus"], subject="fileNr", correction=True, data=error_df[error_df.valueType == "NUMERIC"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [NUMERIC] RM_Ano - C background+focus.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within=["background", "focus"], subject="fileNr", correction=True, data=error_df[error_df.valueType == "BOOL"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [BOOL] RM_Ano - C background+focus.csv")
    pg.rm_anova(dv=VALUE_COLUMN, within=["background", "focus"], subject="fileNr", correction=True, data=times_df[times_df.valueType == "TIME"], detailed=True).to_csv(
                path_or_buf=f"{ANOVA_PATH}/{STUDY_KEY} [TIME] RM_Ano - C background+focus.csv")


# ------- H4 and H8 (+ Further Findings) --------------------
def questionnaire_tests(questionnaire_df: pd.DataFrame):
    os.makedirs(QUESTIONNAIRE_PATH, exist_ok=True)

    health_keys = ["fatigue", "concentration", "motivation", "headache", "dryEyes"]
    # sub dataframes of the values where the first background in the session was either bg3 or bg4 respectively.
    h_before_bg3 = questionnaire_df[questionnaire_df.firstBackground == "BG3"][["fatigueBefore", "concentrationBefore", "motivationBefore", "headacheBefore", "dryEyesBefore"]]
    h_before_bg4 = questionnaire_df[questionnaire_df.firstBackground == "BG4"][["fatigueBefore", "concentrationBefore", "motivationBefore", "headacheBefore", "dryEyesBefore"]]
    h_after_1_bg3 = questionnaire_df[questionnaire_df.firstBackground == "BG3"][["fatigueFirstBG", "concentrationFirstBG", "motivationFirstBG", "headacheFirstBG", "dryEyesFirstBG"]]
    h_after_1_bg4 = questionnaire_df[questionnaire_df.firstBackground == "BG4"][["fatigueFirstBG", "concentrationFirstBG", "motivationFirstBG", "headacheFirstBG", "dryEyesFirstBG"]]
    h_after_2_bg3 = questionnaire_df[questionnaire_df.firstBackground == "BG3"][["fatigueSecondBG", "concentrationSecondBG", "motivationSecondBG", "headacheSecondBG", "dryEyesSecondBG"]]
    h_after_2_bg4 = questionnaire_df[questionnaire_df.firstBackground == "BG4"][["fatigueSecondBG", "concentrationSecondBG", "motivationSecondBG", "headacheSecondBG", "dryEyesSecondBG"]]
    for i, sub_df in enumerate([h_before_bg3, h_before_bg4, h_after_1_bg3, h_after_1_bg4, h_after_2_bg3, h_after_2_bg4]):
        sub_df.columns = health_keys

    # Sort the data according to the background they were coming from.
    h_before_to_bg3 = h_after_1_bg3 - h_before_bg3
    h_before_to_bg3["fileNr"] = questionnaire_df[questionnaire_df.firstBackground == "BG3"].fileNr
    h_before_to_bg3["from"] = "Before"
    h_before_to_bg3["to"] = "BG3"
    h_bg3_to_bg4 = h_after_2_bg3 - h_after_1_bg3
    h_bg3_to_bg4["fileNr"] = questionnaire_df[questionnaire_df.firstBackground == "BG3"].fileNr
    h_bg3_to_bg4["from"] = "BG3"
    h_bg3_to_bg4["to"] = "BG4"
    h_before_to_bg4 = h_after_1_bg4 - h_before_bg4
    h_before_to_bg4["fileNr"] = questionnaire_df[questionnaire_df.firstBackground == "BG4"].fileNr
    h_before_to_bg4["from"] = "Before"
    h_before_to_bg4["to"] = "BG4"
    h_bg4_to_bg3 = h_after_2_bg4 - h_after_1_bg4
    h_bg4_to_bg3["fileNr"] = questionnaire_df[questionnaire_df.firstBackground == "BG4"].fileNr
    h_bg4_to_bg3["from"] = "BG4"
    h_bg4_to_bg3["to"] = "BG3"

    health_to_bg3 = h_before_to_bg3.append(h_bg4_to_bg3)
    health_to_bg4 = h_before_to_bg4.append(h_bg3_to_bg4)
    health_df = health_to_bg3.append(health_to_bg4)

    for key in health_keys:
        pg.friedman(dv=key, within="to", subject="fileNr", data=health_df).to_csv(
                   path_or_buf=f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} Fried - h-{key} delta (to).csv")

    perception_keys = [f"generalQuestion{i}" for i in range(1, 4)] + [f"backgroundPerception{i}" for i in range(1, 3)]
    perception_first_bg3 = questionnaire_df[questionnaire_df.firstBackground == "BG3"][["fileNr"] + [f"generalQuestion{i}FirstBG" for i in range(1, 4)] + [f"backgroundPerception{i}FirstBG" for i in range(1, 3)]]
    perception_first_bg4 = questionnaire_df[questionnaire_df.firstBackground == "BG4"][["fileNr"] + [f"generalQuestion{i}FirstBG" for i in range(1, 4)] + [f"backgroundPerception{i}FirstBG" for i in range(1, 3)]]
    perception_second_bg3 = questionnaire_df[questionnaire_df.firstBackground == "BG3"][["fileNr"] + [f"generalQuestion{i}SecondBG" for i in range(1, 4)] + [f"backgroundPerception{i}SecondBG" for i in range(1, 3)]]
    perception_second_bg4 = questionnaire_df[questionnaire_df.firstBackground == "BG4"][["fileNr"] + [f"generalQuestion{i}SecondBG" for i in range(1, 4)] + [f"backgroundPerception{i}SecondBG" for i in range(1, 3)]]
    for sub_df in [perception_first_bg3, perception_first_bg4, perception_second_bg3, perception_second_bg4]:
        sub_df.columns = ["fileNr"] + perception_keys

    perception_bg3 = perception_first_bg3.append(perception_second_bg4)
    perception_bg4 = perception_first_bg4.append(perception_second_bg3)
    perception_df = perception_bg3.append(perception_bg4)
    perception_df["to"] = ["BG3"] * questionnaire_df.shape[0] + ["BG4"] * questionnaire_df.shape[0]

    for key in perception_keys:
        pg.friedman(dv=key, within="to", subject="fileNr", data=perception_df).to_csv(
                   path_or_buf=f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} Fried - {key} (to).csv")

    sub_df = questionnaire_df.drop(labels=["sex", "fileNr"], axis=1)
    sub_df.mean().to_csv(f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} Mean.csv", header=True)
    sub_df.std().to_csv(f"{QUESTIONNAIRE_PATH}/{STUDY_KEY} STD.csv", header=True)


if __name__ == '__main__':
    os.makedirs(OUTPUT_PATH, exist_ok=True)

    error_df = calc_values()
    error_df = error_df[~error_df.questionType.isin(["-", QUESTION_TYPES_CALCULATED[2], QUESTION_TYPES_CALCULATED[3], QUESTION_TYPES[5]])]
    error_df[VALUE_COLUMN] = error_df.apply(lambda row: (1 if row[VALUE_COLUMN] else 0) if row["valueType"] is "BOOL" else row[VALUE_COLUMN], axis=1)
    error_df = error_df.astype({VALUE_COLUMN: np.float64})

    times_df = calc_time_values()
    times_df = times_df[times_df.valueDescription.isin(["SolvingDuration", "Solving-#Toggles"])]
    times_df = times_df.astype({VALUE_COLUMN: np.float64})

    questionnaire_df = get_questionnaire_data()

    check_preliminaries(error_df, "E")
    check_preliminaries(times_df, "T")
    statistical_tests(error_df, times_df)
    questionnaire_tests(questionnaire_df)
